/* $Id: unsteadynavierstokes.h,v 1.5 1999/08/25 19:37:43 knepley Exp $ */

/* Main include file for the Navier-Stokes flow test problem */

#ifndef __UNSTEADY_NAVIER_STOKES_H
#define __UNSTEADY_NAVIER_STOKES_H

#include "petscts.h"
#include "gsolver.h"
#include "src/snes/impls/ml/mlSNES.h"

extern int UNSTEADY_NAVIERSTOKES_COOKIE;
extern int UNSTEADY_NAVIERSTOKES_ComputeField;

enum {Stage_Setup, Stage_GridRefine, NumStages};
int *UserStages;

/* This structure holds information about physical properties */
typedef struct {
  double      nu;        /* The kinematic viscosity */
  double      fluidSG;   /* The specific gravity of the fluid */
  PetscScalar frequency; /* The frequency of the exact solution */
} PhysicalContext;

struct _UnsteadyNavierStokesContext {
  PETSCHEADER(int)
  Grid                   grid;              /* The problem grid */

  MeshBoundary2D         boundaryCtx;       /* Describes the domain boundary */
  MeshGeometryContext    geometryCtx;       /* Describes the domain geometry */
  PhysicalContext        physicalCtx;       /* Describes the fluid properties */

  int                    RofS;              /* The rate-of-strain operator */
  PetscTruth             useML;             /* Flag for ML preconditioning */
  PetscTruth             useSchur;          /* Flag for Schur complement preconditioning */

  GTS                    ts;                /* The integrator context */
  GVec                   u;                 /* The velocity solution vector */
  GVec                   p;                 /* The pressure solution vector */
  GVec                   uExact;            /* The exact velocity solution vector */
  GVec                   pExact;            /* The exact pressure solution vector */

  int                    dim;               /* The problem dimension */
  PetscTruth             linear;            /* Using linear elements (as opposed to quadratic) */
  int                    refineStep;        /* The step to start refining the mesh */
  int                    numLoops;          /* The number of systems to solve */

  PetscViewer            mathViewer;        /* Connection to Mathematica */
  PetscViewer            nonlinearViewer;   /* A viewer for analysis of the nonlinear iteration */
  PetscDrawHG            nonlinearHist;     /* A histogram of the number of nonlinear iterates for each time step */
  PetscViewer            linearViewer;      /* A viewer for analysis of the linear iteration */
  PetscDrawHG            linearHist;        /* A histogram of the number of linear iterates for each Newton step */
  PetscViewer            totalLinearViewer; /* A viewer for analysis of the linear iteration */
  PetscDrawHG            totalLinearHist;   /* A histogram of the number of total linear iterates in a nonlinear solve */
  PetscViewer            innerLinearViewer; /* A viewer for analysis of preconditioner iterations */
  PetscDrawHG            innerLinearHist;   /* A histogram of the number of linear iterates in a solve with A */
  PetscViewer            schurLinearViewer; /* A viewer for analysis of Schur complement iterations */
  PetscDrawHG            schurLinearHist;   /* A histogram of the number of linear iterates in a solve with B^T A^{-1} B */
};
typedef struct _UnsteadyNavierStokesContext *UnsteadyNavierStokesContext;

/* Creation functions */
EXTERN int UnsteadyNavierStokesInitializePackage(char *);
EXTERN int UnsteadyNavierStokesContextCreate(MPI_Comm, UnsteadyNavierStokesContext *);
EXTERN int UnsteadyNavierStokesSetup(UnsteadyNavierStokesContext);
EXTERN int UnsteadyNavierStokesCreateGrid(UnsteadyNavierStokesContext);
EXTERN int UnsteadyNavierStokesRefineGrid(UnsteadyNavierStokesContext);
EXTERN int UnsteadyNavierStokesCreateStructures(UnsteadyNavierStokesContext);

/* Setup functions */
EXTERN int UnsteadyNavierStokesSetupGrid(UnsteadyNavierStokesContext);
EXTERN int UnsteadyNavierStokesSetupRhsFunction(Grid, UnsteadyNavierStokesContext);
EXTERN int UnsteadyNavierStokesSetupBC(Grid, UnsteadyNavierStokesContext);
EXTERN int UnsteadyNavierStokesSetupSNES(SNES, UnsteadyNavierStokesContext);
EXTERN int UnsteadyNavierStokesSetupKSP(KSP, UnsteadyNavierStokesContext);
EXTERN int UnsteadyNavierStokesSetupPC(PC, UnsteadyNavierStokesContext);
EXTERN int UnsteadyNavierStokesSetupStructures(UnsteadyNavierStokesContext);

/* Cleanup functions */
EXTERN int UnsteadyNavierStokesDestroyStructures(UnsteadyNavierStokesContext);
EXTERN int UnsteadyNavierStokesContextDestroy(UnsteadyNavierStokesContext);

/* Checking functions */
EXTERN int MatCheckSymmetry(Mat);
EXTERN int UnsteadyNavierStokesCheckSolution(UnsteadyNavierStokesContext, double, GVec, GVec, const char []);

/* Callback functions */
EXTERN int UnsteadyNavierStokesSolutionFunction(Vec, void *);
EXTERN int VelocitySolutionFunction(int, int, double *, double *, double *, PetscScalar *, void*);
EXTERN int VelocitySolutionTimeDerivative(int, int, double *, double *, double *, PetscScalar *, void*);
EXTERN int PressureSolutionFunction(int, int, double *, double *, double *, PetscScalar *, void*);
EXTERN int VelocityRhsFunction(int, int, double *, double *, double *, PetscScalar *, void*);
EXTERN int UnsteadyNavierStokesUpdateGTS(TS, double, double *);

/* Operator functions */
EXTERN int ConvectionOperator(int, int, double *, double *, double *, int, PetscScalar **, PetscScalar *, void *);
EXTERN int LinearConvectionOperator(Discretization, Discretization, int, int, int, int, int, double *,
                                    PetscScalar, PetscScalar *, PetscScalar *, void *);
EXTERN int RateOfStrainTensor(Discretization, Discretization, int, int, int, int, int, double *,
                              PetscScalar, PetscScalar *, PetscScalar *, void *);

/* Analysis functions */
EXTERN int UnsteadyNavierStokesCreateHists(UnsteadyNavierStokesContext);
EXTERN int UnsteadyNavierStokesResetHists(UnsteadyNavierStokesContext, PetscTruth);
EXTERN int UnsteadyNavierStokesDestroyHists(UnsteadyNavierStokesContext);
EXTERN int UnsteadyNavierStokesUpdateLinearHists(SLES, UnsteadyNavierStokesContext);
EXTERN int UnsteadyNavierStokesUpdateGSNES(GSNES, int);
EXTERN int UnsteadyNavierStokesAnalyzeSolvers(UnsteadyNavierStokesContext);

/* Solver functions */
EXTERN int UnsteadyNavierStokesSolve(UnsteadyNavierStokesContext, double *);
EXTERN int UnsteadyNavierStokesComputeFlowField(UnsteadyNavierStokesContext);

#endif /* __NAVIER_STOKES_H */
